/*
 Illinois Open Source License

 University of Illinois/NCSA
 Open Source License

 Copyright © 2009,    University of Illinois.  All rights reserved.

 Developed by:
 Innovative Systems Lab
 National Center for Supercomputing Applications
 http://www.ncsa.uiuc.edu/AboutUs/Directorates/ISL.html

 Permission is hereby granted, free of charge, to any person obtaining a copy of this 
 software and associated documentation files (the "Software"), to deal with the Software 
 without restriction, including without limitation the rights to use, copy, modify, merge, 
 publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
 persons to whom the Software is furnished to do so, subject to the following conditions:
 
 * Redistributions of source code must retain the above copyright notice, this list of 
   conditions and the following disclaimers.
 
 * Redistributions in binary form must reproduce the above copyright notice, this list of 
   conditions and the following disclaimers in the documentation and/or other materials 
   provided with the distribution.
 
 * Neither the names of Innovative Systems Lab and National Center for Supercomputing 
   Applications, nor the names of its contributors may be used to endorse or promote 
   products derived from this Software without specific prior written permission.
 
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
   INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
   PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE 
   LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
   OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
   DEALINGS WITH THE SOFTWARE.
 */

#include <sys/time.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <libspe2.h>
#include <pthread.h>

#include "model.h"
#include "spu/compute_task.h"
#include <pputask.h>

compute_task_t compute_task __attribute__ ((aligned (128)));
load_task_t load_task __attribute__ ((aligned (128)));

int
doComputeWaterfall(struct cartesian *data1, int n1, struct cartesian *data2, struct cartesian2* data2_v, int n2, 
		   int doSelf, long long data_bins[][MAXBINS], int njk, int flag, int speindex)
{ 
    
    compute_task_t * task = & compute_task;
    task->common.cmd = COMPUTE_TASK;
    task->common.size = sizeof(compute_task);
    task->data1 = (char*)data1;
    task->n1 = n1;
    task->data2 = (char*)data2;
    task->data2_v = (char*) data2_v;
    task->n2 = n2;
    task->doSelf = doSelf;
    task->data_bins =(long long**) data_bins;
    task->njk = njk;
    task->flag=flag;
    ppu_task_spu_run((task_t*)task, speindex);
    
    return 0;
}



int
upload_binbs(double* binb, int nbins)
{
    int i;
    load_task_t* task = &load_task;
    task->common.cmd = LOAD_TASK;
    task->common.size = sizeof(load_task_t);
    task->binb = binb;
    task->nbins = nbins;
    ppu_task_run((task_t*)task);
    
    return 0;
}


int
doComputeWaterfall_cpu(struct cartesian *data1, int n1, struct cartesian *data2, int n2, int doSelf, long long data_bins[][MAXBINS], int nbins, double *binb, int njk)
{
    int i, j, l;
    for(i=0; i<((doSelf) ? n1-1 : n1); i++)
        {
            register double xi = data1[i].x;
            register double yi = data1[i].y;
            register double zi = data1[i].z;
            register int jk = data1[i].jk;
            for(j = ((doSelf) ? i+1 : 0); j<n2; j++)
                {
                    register int indx;
                    register double dot = xi * data2[j].x + yi * data2[j].y + zi * data2[j].z;
                    //put in correct bin
                    register int k = nbins;
                    if (dot >= binb[0])
                        indx = 0;
                    else
                        {
                            while (dot > binb[k]) k--;
                            indx = k+1;
                        }
                    //update all samples
                    for(l=0; l<njk; l++)
                        if(l!=jk)
                            data_bins[l][indx] +=1;
                }
        }

    return 0;
}







