/*
 Illinois Open Source License

 University of Illinois/NCSA
 Open Source License

 Copyright © 2009,    University of Illinois.  All rights reserved.

 Developed by:
 Innovative Systems Lab
 National Center for Supercomputing Applications
 http://www.ncsa.uiuc.edu/AboutUs/Directorates/ISL.html

 Permission is hereby granted, free of charge, to any person obtaining a copy of this 
 software and associated documentation files (the "Software"), to deal with the Software 
 without restriction, including without limitation the rights to use, copy, modify, merge, 
 publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
 persons to whom the Software is furnished to do so, subject to the following conditions:
 
 * Redistributions of source code must retain the above copyright notice, this list of 
   conditions and the following disclaimers.
 
 * Redistributions in binary form must reproduce the above copyright notice, this list of 
   conditions and the following disclaimers in the documentation and/or other materials 
   provided with the distribution.
 
 * Neither the names of Innovative Systems Lab and National Center for Supercomputing 
   Applications, nor the names of its contributors may be used to endorse or promote 
   products derived from this Software without specific prior written permission.
 
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
   INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
   PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE 
   LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
   OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
   DEALINGS WITH THE SOFTWARE.
 */

#include <stdio.h>
#include <string.h>
#include <spu_mfcio.h>
 
#include "../model.h"
#include "compute_task.h"
#include "spu_misc.h"
#include <sputask.h>


void convert_back(vector float* src, vector float * dst);
extern vector unsigned char shuffle_table1[];
extern vector unsigned char shuffle_table2[];
extern vector unsigned char double_shuf_table1[];
extern vector unsigned char double_shuf_table2[];
extern vector unsigned char double_shuf_table3[];
extern vector unsigned char double_shuf_table4[];

#define BLOCK 512

struct cartesian data1[BLOCK];
struct cartesian data2[BLOCK];

static vector unsigned int index_array[BLOCK/4];

long long data_bins[MAXNJK][MAXBINS] __attribute__ ((aligned (128)));
long long temp_data_bins[MAXBINS] __attribute__ ((aligned (128)));

extern double binb_d[];
extern float binb[];
extern vector float binb_v[];
extern vector double binb_dv[];
extern int nbins;
int njk;



extern vector unsigned short large_table[];

vector unsigned short
compute_subtotal(int index_array_count)
{
    int i;
    vector unsigned short subtotal = (vector unsigned short){0,0,0,0, 0,0,0,0};

    for (i=0;i < ROUNDDOWN_16(index_array_count); i+=16){

#define URF 4 /*unroll factor*/	
	vector unsigned int * v = &index_array[i];
	vector unsigned int t0[URF],t1[URF],t2[URF],t3[URF],t4[URF],t5[URF];
	int indx0[URF],indx1[URF],indx2[URF],indx3[URF];
	vector unsigned short entry0[URF], entry1[URF], entry2[URF], entry3[URF];
	vector unsigned short tsum0[URF], tsum1[URF];
	
#define ROTATE_AND_COMPOSE(idx) \
	t0[idx] = spu_rl(v[idx*4], 9); \
	t1[idx] = spu_rl(v[idx*4 + 1], 6); \
	t2[idx] = spu_rl(v[idx*4 + 2], 3); \
	t3[idx] = spu_or(t0[idx], t1[idx]); \
	t4[idx] = spu_or(t2[idx], v[idx*4+3]); \
	t5[idx] = spu_or(t3[idx], t4[idx]);  \
	indx0[idx] = spu_extract(t5[idx], 0); \
	indx1[idx] = spu_extract(t5[idx], 1); \
	indx2[idx] = spu_extract(t5[idx], 2); \
	indx3[idx] = spu_extract(t5[idx], 3); \
	entry0[idx] = large_table[indx0[idx]]; \
	entry1[idx] = large_table[indx1[idx]]; \
	entry2[idx] = large_table[indx2[idx]]; \
	entry3[idx] = large_table[indx3[idx]];	 \
	tsum0[idx] = spu_add(entry0[idx], entry1[idx]); \
	tsum1[idx] = spu_add(entry2[idx], entry3[idx]); \
	subtotal = spu_add(subtotal, tsum0[idx]); \
	subtotal = spu_add(subtotal, tsum1[idx]);
	
	ROTATE_AND_COMPOSE(0);
	ROTATE_AND_COMPOSE(1);
	ROTATE_AND_COMPOSE(2);
	ROTATE_AND_COMPOSE(3);

    }

    for (;i < ROUNDDOWN_4(index_array_count); i+=4){
	
	vector unsigned int * v = &index_array[i];
	vector unsigned int t0[URF],t1[URF],t2[URF],t3[URF],t4[URF],t5[URF];
	int indx0[URF],indx1[URF],indx2[URF],indx3[URF];
	vector unsigned short entry0[URF], entry1[URF], entry2[URF], entry3[URF];
	vector unsigned short tsum0[URF], tsum1[URF];
		
	ROTATE_AND_COMPOSE(0);

    }



    for (;i < index_array_count; i++){
	
	int vl;
	for (vl =0;vl < 4;vl++){
	    vector unsigned short subtotal_addition[]={
		{1,0,0,0, 0,0,0,0},
		{0,1,0,0, 0,0,0,0},
		{0,0,1,0, 0,0,0,0},
		{0,0,0,1, 0,0,0,0},
		{0,0,0,0, 1,0,0,0},
		{0,0,0,0, 0,1,0,0},
		{0,0,0,0, 0,0,1,0},
		{0,0,0,0, 0,0,0,1}
	    };
	    
	    unsigned int indx = spu_extract(index_array[i], vl);
	    subtotal = spu_add(subtotal, subtotal_addition[indx]);
	}
	
	
    }   
    
    return subtotal;
    
}




int prev_count =0;
vector float dot_save;
vector double dot_save1;
vector double dot_save2;
int index_array2_count = 0;
extern vector unsigned int index_array2[];
void 
partial_compute_dot(int jk)
{
    int indx;
    int vl;
    for (vl = 0;vl < prev_count; vl++){
	double dot;
	if (vl <2){
	    dot = spu_extract(dot_save1, vl);
	}else{
	    dot = spu_extract(dot_save2, vl -2);
	}
	
	int k = nbins;
	if (dot >= binb_d[0])
	    indx = 0;
	else{
	    while (dot > binb_d[k]){
		k--;
	    }
	    indx = k+1;
	}
	if (indx < 24){
	    temp_data_bins[indx] +=1;
	    if (jk >=0){
		data_bins[jk][indx] -=1;
	    }
	}
    }	
    
    prev_count = 0;
}


void 
test_compute_smallp(vector double dot_v1, vector double dot_v2,  vector unsigned int prev_dotgather, int jk)
{
    
    vector unsigned char cmp_shuf_v={
	0x80,0x80,0x80,0x80,  0x80,0x80,0x80,0x80, 0,1,2,3, 8,9,10,11
    };
    vector unsigned long long dotcmp1 = spu_cmpgt(dot_v1, binb_dv[nbins-15]);
    vector unsigned long long dotcmp2 = spu_cmpgt(dot_v2, binb_dv[nbins-15]);
    vector unsigned int adjusted_dotcmp1 = (vector unsigned int)spu_shuffle(dotcmp1, dotcmp1, cmp_shuf_v);
    vector unsigned int adjusted_dotcmp2 = (vector unsigned int)spu_shuffle(dotcmp2, dotcmp2, cmp_shuf_v);
    
    vector unsigned int dotgather1 = spu_gather(adjusted_dotcmp1);
    vector unsigned int dotgather2 = spu_gather(adjusted_dotcmp2);
    
    vector unsigned int cmpmask = spu_splats((unsigned int)0x01);      
    vector unsigned char shuf_1, shuf_2, shuf_3, shuf_4;
    vector double temp_dot1, temp_dot2;
    vector unsigned long long tmp_cmps_a[8], tmp_cmps_b[8];
    vector unsigned int cmps[8];
    vector unsigned int t0,t1,t2,t3,t4,t5, t6;    
    vector unsigned char cmp_merge_v={
	0,1,2,3,  8,9,10,11, 16,17,18,19, 24,25,26,27
    };
    
    if (spu_extract(dotgather1,0) == 0 && spu_extract(dotgather2, 0)== 0){
	//if (0){
	dotcmp1 = spu_cmpgt(dot_v1, binb_dv[nbins-7]);
	dotcmp2 = spu_cmpgt(dot_v2, binb_dv[nbins-7]);
	adjusted_dotcmp1 = (vector unsigned int)spu_shuffle(dotcmp1, dotcmp1, cmp_shuf_v);
	adjusted_dotcmp2 = (vector unsigned int)spu_shuffle(dotcmp2, dotcmp2, cmp_shuf_v);
	dotgather1 = spu_gather(adjusted_dotcmp1);
	dotgather2 = spu_gather(adjusted_dotcmp2);
	
	int temp_count = prev_count;
	int curr = spu_extract(dotgather1, 0);
	int currbits = spu_extract(spu_cntb((vector unsigned char)dotgather1), 3);
	int index = temp_count* 4 + curr;
	temp_count += currbits;

#if 0	
	spu_log("prev_count=%d, (dot1) curr=%d, currbits=%d, (dot2) curr=%d, currbits=%d\n", 
		prev_count,
		spu_extract(dotgather1, 0),
		spu_extract(spu_cntb((vector unsigned char)dotgather1), 3),
		spu_extract(dotgather2, 0),
		spu_extract(spu_cntb((vector unsigned char)dotgather2), 3));
		
		

	PRINTVD(dot_save1);
	PRINTVD(dot_save2);
	PRINTVI(dotgather1);
	PRINTVI(dotgather2);
	PRINTVD(dot_v1);
	PRINTVD(dot_v2);
	printf("\n");

#endif
	
	if (temp_count >=4 ){
	    
	    
	    //construct two full double vector
	    shuf_3 = double_shuf_table3[index];      
	    shuf_4 = double_shuf_table4[index];            
	    temp_dot1 = spu_shuffle(dot_save1, dot_v1, shuf_3);
	    temp_dot2 = spu_shuffle(dot_save2, dot_v1, shuf_4);
	    
	    
	    //do computation for the two vector double
	    tmp_cmps_a[0]= spu_cmpgt(temp_dot1, binb_dv[nbins-8]);  
	    tmp_cmps_b[0] =spu_cmpgt(temp_dot2, binb_dv[nbins-8]);   
	    cmps[0] = (vector unsigned int)spu_shuffle(tmp_cmps_a[0], tmp_cmps_b[0], cmp_merge_v);           
	    tmp_cmps_a[1]= spu_cmpgt(temp_dot1, binb_dv[nbins-9]);   
	    tmp_cmps_b[1]= spu_cmpgt(temp_dot2, binb_dv[nbins-9]);  
	    cmps[1] = (vector unsigned int)spu_shuffle(tmp_cmps_a[1], tmp_cmps_b[1], cmp_merge_v); 
	    tmp_cmps_a[2]= spu_cmpgt(temp_dot1, binb_dv[nbins-10]);  
	    tmp_cmps_b[2]= spu_cmpgt(temp_dot2, binb_dv[nbins-10]);  
	    cmps[2] = (vector unsigned int)spu_shuffle(tmp_cmps_a[2], tmp_cmps_b[2], cmp_merge_v); 
	    tmp_cmps_a[3]= spu_cmpgt(temp_dot1, binb_dv[nbins-11]);  
	    tmp_cmps_b[3]= spu_cmpgt(temp_dot2, binb_dv[nbins-11]);  
	    cmps[3] = (vector unsigned int)spu_shuffle(tmp_cmps_a[3], tmp_cmps_b[3], cmp_merge_v); 
	    tmp_cmps_a[4]= spu_cmpgt(temp_dot1, binb_dv[nbins-12]);  
	    tmp_cmps_b[4]= spu_cmpgt(temp_dot2, binb_dv[nbins-12]); 
	    cmps[4] = (vector unsigned int)spu_shuffle(tmp_cmps_a[4], tmp_cmps_b[4], cmp_merge_v); 
	    tmp_cmps_a[5]= spu_cmpgt(temp_dot1, binb_dv[nbins-13]);  
	    tmp_cmps_b[5]= spu_cmpgt(temp_dot2, binb_dv[nbins-13]);  
	    cmps[5] = (vector unsigned int)spu_shuffle(tmp_cmps_a[5], tmp_cmps_b[5], cmp_merge_v); 
	    tmp_cmps_a[6]= spu_cmpgt(temp_dot1, binb_dv[nbins-14]);  
	    tmp_cmps_b[6]= spu_cmpgt(temp_dot2, binb_dv[nbins-14]);  
	    cmps[6] = (vector unsigned int)spu_shuffle(tmp_cmps_a[6], tmp_cmps_b[6], cmp_merge_v); 
	    tmp_cmps_a[7]= spu_cmpgt(temp_dot1, binb_dv[nbins-15]);  
	    tmp_cmps_b[7]= spu_cmpgt(temp_dot2, binb_dv[nbins-15]); 
	    cmps[7] = (vector unsigned int)spu_shuffle(tmp_cmps_a[7], tmp_cmps_b[7], cmp_merge_v); 
	    cmps[0] = spu_and(cmps[0], cmpmask); 
	    cmps[1] = spu_and(cmps[1], cmpmask); 
	    cmps[2] = spu_and(cmps[2], cmpmask); 
	    cmps[3] = spu_and(cmps[3], cmpmask); 
	    cmps[4] = spu_and(cmps[4], cmpmask); 
	    cmps[5] = spu_and(cmps[5], cmpmask); 
	    cmps[6] = spu_and(cmps[6], cmpmask); 
	    cmps[7] = spu_and(cmps[7], cmpmask); 
	    t0 = spu_add(cmps[0], cmps[1]);  
	    t1 = spu_add(cmps[2], cmps[3]);  
	    t2 = spu_add(cmps[4], cmps[5]);  
	    t3 = spu_add(cmps[6], cmps[7]);  
	    t4 = spu_add(t0, t1);  
	    t5 = spu_add(t2, t3);  
	    t6 = spu_add(t4, t5);  
	    index_array2[index_array2_count] = t6;	    
	    index_array2_count++;
	    temp_count -=4;      
	}
	
	shuf_1 = double_shuf_table1[index];
	shuf_2 = double_shuf_table2[index];
	
	dot_save1 = spu_shuffle(dot_save1, dot_v1, shuf_1);
	dot_save2 = spu_shuffle(dot_save2, dot_v1, shuf_2);
	

	curr = spu_extract(dotgather2, 0);
	currbits = spu_extract(spu_cntb((vector unsigned char)dotgather2), 3);
	index = temp_count* 4 + curr;
	temp_count += currbits;
	if (temp_count >=4 ){
	    
	    //construct two full double vector
	    shuf_3 = double_shuf_table3[index];      
	    shuf_4 = double_shuf_table4[index];            
	    temp_dot1 = spu_shuffle(dot_save1, dot_v2, shuf_3);
	    temp_dot2 = spu_shuffle(dot_save2, dot_v2, shuf_4);
	 
	    
	    //do computation for the two vector double
	    tmp_cmps_a[0]= spu_cmpgt(temp_dot1, binb_dv[nbins-8]);  
	    tmp_cmps_b[0] =spu_cmpgt(temp_dot2, binb_dv[nbins-8]);   
	    cmps[0] = (vector unsigned int)spu_shuffle(tmp_cmps_a[0], tmp_cmps_b[0], cmp_merge_v);           
	    tmp_cmps_a[1]= spu_cmpgt(temp_dot1, binb_dv[nbins-9]);   
	    tmp_cmps_b[1]= spu_cmpgt(temp_dot2, binb_dv[nbins-9]);  
	    cmps[1] = (vector unsigned int)spu_shuffle(tmp_cmps_a[1], tmp_cmps_b[1], cmp_merge_v); 
	    tmp_cmps_a[2]= spu_cmpgt(temp_dot1, binb_dv[nbins-10]);  
	    tmp_cmps_b[2]= spu_cmpgt(temp_dot2, binb_dv[nbins-10]);  
	    cmps[2] = (vector unsigned int)spu_shuffle(tmp_cmps_a[2], tmp_cmps_b[2], cmp_merge_v); 
	    tmp_cmps_a[3]= spu_cmpgt(temp_dot1, binb_dv[nbins-11]);  
	    tmp_cmps_b[3]= spu_cmpgt(temp_dot2, binb_dv[nbins-11]);  
	    cmps[3] = (vector unsigned int)spu_shuffle(tmp_cmps_a[3], tmp_cmps_b[3], cmp_merge_v); 
	    tmp_cmps_a[4]= spu_cmpgt(temp_dot1, binb_dv[nbins-12]);  
	    tmp_cmps_b[4]= spu_cmpgt(temp_dot2, binb_dv[nbins-12]); 
	    cmps[4] = (vector unsigned int)spu_shuffle(tmp_cmps_a[4], tmp_cmps_b[4], cmp_merge_v); 
	    tmp_cmps_a[5]= spu_cmpgt(temp_dot1, binb_dv[nbins-13]);  
	    tmp_cmps_b[5]= spu_cmpgt(temp_dot2, binb_dv[nbins-13]);  
	    cmps[5] = (vector unsigned int)spu_shuffle(tmp_cmps_a[5], tmp_cmps_b[5], cmp_merge_v); 
	    tmp_cmps_a[6]= spu_cmpgt(temp_dot1, binb_dv[nbins-14]);  
	    tmp_cmps_b[6]= spu_cmpgt(temp_dot2, binb_dv[nbins-14]);  
	    cmps[6] = (vector unsigned int)spu_shuffle(tmp_cmps_a[6], tmp_cmps_b[6], cmp_merge_v); 
	    tmp_cmps_a[7]= spu_cmpgt(temp_dot1, binb_dv[nbins-15]);  
	    tmp_cmps_b[7]= spu_cmpgt(temp_dot2, binb_dv[nbins-15]); 
	    cmps[7] = (vector unsigned int)spu_shuffle(tmp_cmps_a[7], tmp_cmps_b[7], cmp_merge_v); 
	    cmps[0] = spu_and(cmps[0], cmpmask); 
	    cmps[1] = spu_and(cmps[1], cmpmask); 
	    cmps[2] = spu_and(cmps[2], cmpmask); 
	    cmps[3] = spu_and(cmps[3], cmpmask); 
	    cmps[4] = spu_and(cmps[4], cmpmask); 
	    cmps[5] = spu_and(cmps[5], cmpmask); 
	    cmps[6] = spu_and(cmps[6], cmpmask); 
	    cmps[7] = spu_and(cmps[7], cmpmask); 
	    t0 = spu_add(cmps[0], cmps[1]);  
	    t1 = spu_add(cmps[2], cmps[3]);  
	    t2 = spu_add(cmps[4], cmps[5]);  
	    t3 = spu_add(cmps[6], cmps[7]);  
	    t4 = spu_add(t0, t1);  
	    t5 = spu_add(t2, t3);  
	    t6 = spu_add(t4, t5);  
	    index_array2[index_array2_count] = t6;	    
	    index_array2_count++;
	    temp_count -=4;      
	    
	}
	shuf_1 = double_shuf_table1[index];
	shuf_2 = double_shuf_table2[index];
	
	dot_save1 = spu_shuffle(dot_save1, dot_v2, shuf_1);
	dot_save2 = spu_shuffle(dot_save2, dot_v2, shuf_2);
	
	prev_count =temp_count;
	
    }else{
	int indx;
	int vl;
	for (vl = 0;vl < 4; vl++){
	    double dot;
	    if (vl <2){
		dot = spu_extract(dot_v1, vl);
	    }else{
		dot = spu_extract(dot_v2, vl -2);
	    }
	    int k = nbins-7;
	    if (dot < binb_d[nbins-7]){
		continue;
	    }
	    if (dot >= binb_d[0])
		indx = 0;
	    else{
		while (dot > binb_d[k]){
		    k--;
		}
		indx = k+1;
	    }
	    temp_data_bins[indx] +=1;
	    if (jk >=0){
		data_bins[jk][indx] -=1;
	    }
	}		    		    
    }
}


void 
compute_smallp(vector double dot_v1, vector double dot_v2,  vector unsigned int prev_dotgather, int jk)
{
#if 0 
    vector unsigned int dotcmp = spu_cmpgt(dot_v, binb_v[nbins-15]);
    vector unsigned int dotgather = spu_gather(dotcmp);
    vector unsigned int cmps[16];
    vector unsigned int cmpmask = spu_splats((unsigned int)0x01);
#endif

    //if (spu_extract(dotgather,0) == 0){
    if (0){
#if 0
	int temp_count = prev_count;
	int curr = spu_extract(prev_dotgather, 0);
	int currbits = spu_extract(spu_cntb((vector unsigned char)prev_dotgather), 3);
	int index = temp_count* 16 + curr;
	temp_count += currbits;
	if (temp_count >=4 ){
	    
	    vector unsigned char shuffle_v2 = shuffle_table2[index];
	    vector float temp_dot = spu_shuffle(dot_save, dot_v, shuffle_v2);

	    vector unsigned int t7, t8, t9, t10, t11, t12, t13,t14;
	    cmps[8]= spu_cmpgt(temp_dot, binb_v[nbins-8]);
	    cmps[9]= spu_cmpgt(temp_dot, binb_v[nbins-9]);
	    cmps[10]= spu_cmpgt(temp_dot, binb_v[nbins-10]);
	    cmps[11]= spu_cmpgt(temp_dot, binb_v[nbins-11]);
	    cmps[12]= spu_cmpgt(temp_dot, binb_v[nbins-12]);
	    cmps[13]= spu_cmpgt(temp_dot, binb_v[nbins-13]);
	    cmps[14]= spu_cmpgt(temp_dot, binb_v[nbins-14]);
	    
	    cmps[8] = spu_and(cmps[8], cmpmask);
	    cmps[9] = spu_and(cmps[9], cmpmask);
	    cmps[10] = spu_and(cmps[10], cmpmask);
	    cmps[11] = spu_and(cmps[11], cmpmask);
	    cmps[12] = spu_and(cmps[12], cmpmask);
	    cmps[13] = spu_and(cmps[13], cmpmask);
	    cmps[14] = spu_and(cmps[14], cmpmask);
	    
	    t7 = spu_add(cmps[8], cmps[9]);
	    t8 = spu_add(cmps[10], cmps[11]);
	    t9 = spu_add(cmps[12], cmps[13]);
	    t10 = spu_add(cmps[14], dotcmp);
	    
	    t11 = spu_add(t7, t8);
	    t12 = spu_add(t9, t10);
	    t13 = spu_add(t11, t12);
	    
	    index_array2[index_array2_count] = t13;	    
	    index_array2_count++;
	    temp_count -=4;
	    
	}
	
	vector unsigned char shuffle_v1 = shuffle_table1[index];
	dot_save =spu_shuffle(dot_save, dot_v, shuffle_v1);
	   
	prev_count =temp_count;
#endif

	
    }else{	 
	int indx;
	int vl;
	for (vl = 0;vl < 4; vl++){
	    double dot;
	    if (vl <2){
		dot = spu_extract(dot_v1, vl);
	    }else{
		dot = spu_extract(dot_v2, vl -2);
	    }
	    int k = nbins-7;
	    if (dot < binb_d[nbins-7]){
		continue;
	    }
	    if (dot >= binb_d[0])
		indx = 0;
	    else{
		while (dot > binb_d[k]){
		    k--;
		}
		indx = k+1;
	    }
	    temp_data_bins[indx] +=1;
	    if (jk >=0){
		data_bins[jk][indx] -=1;
	    }
	}		    		
    }    
}

vector unsigned short compute_subtotal2(int index_array_count);
void
block_compute_scalar(int blocksize_i, int blocksize_j, struct cartesian* d1, struct cartesian* d2)
{
    int i, j;
    for (i =0;i < blocksize_i; i++){
	for (j=0; j < blocksize_j; j++){
	    
	    struct cartesian myd2;
	    if (j % 2 ==0){
		double* p = (double*)&d2[j];
		myd2.x = p[0];
		myd2.y = p[2];
		myd2.z = p[4];
		myd2.jk = *(unsigned int*)(p+6);
	    }else{
		double* p = (double*)&d2[j-1];
		myd2.x = p[1];
		myd2.y = p[3];
		myd2.z = p[5];
		myd2.jk = *(unsigned int*)(p+7);
		
	    }
	    compute_one_scalar(&d1[i], &myd2);
	    
	}
    }
    
    
}
void
block_compute(int blocksize_i, int blocksize_j, struct cartesian* d1, struct cartesian* d2)
{
    vector unsigned int cmpmask = spu_splats((unsigned int)0x01);
    vector unsigned int i_zero_v =spu_splats((unsigned int)0);
    vector unsigned int i_seven_v =spu_splats((unsigned int)7);

    int ii, jj, k, l;
    for(ii=0; ii<blocksize_i; ii++){
	int jk = d1[ii].jk;					
	vector double x_ii = spu_splats(d1[ii].x);
	vector double y_ii = spu_splats(d1[ii].y);
	vector double z_ii = spu_splats(d1[ii].z);
	
	int index_array_count =0;
	index_array2_count = 0;
	vector double dot_v[16];
	vector unsigned long long tmp_cmps_a[8], tmp_cmps_b[8];
	vector unsigned int cmps[8];
	vector unsigned int dotgather[8];
	vector unsigned int adjusted_index[8];
	vector unsigned int t0,t1,t2,t3,t4,t5, t6;

	vector unsigned char cmp_shuf_v={
	    0,1,2,3,  8,9,10,11, 16,17,18,19, 24,25,26,27
	};
	vector unsigned int local_sel_v;
	int idx1, idx2;

	for(jj = 0; jj<ROUNDDOWN_32(blocksize_j); jj+= 32){
	    vector double* m = (vector double*)(&d2[jj]);
	    
#define COMPUTE_DOTPROD(idx) \
	    idx1 = 2*idx; \
	    idx2 = 2*idx+1; \
	    dot_v[idx1] = spu_mul(x_ii, m[idx1*4]);  \
	    dot_v[idx1] = spu_madd(y_ii, m[idx1*4+1], dot_v[idx1]); \
	    dot_v[idx1] = spu_madd(z_ii, m[idx1*4+2], dot_v[idx1]);  \
	    dot_v[idx2] = spu_mul(x_ii, m[idx2*4]);  \
	    dot_v[idx2] = spu_madd(y_ii, m[idx2*4+1], dot_v[idx2]); \
	    dot_v[idx2] = spu_madd(z_ii, m[idx2*4+2], dot_v[idx2]); 	    
	    
	    
	    COMPUTE_DOTPROD(0);
	    COMPUTE_DOTPROD(1);
	    COMPUTE_DOTPROD(2);
	    COMPUTE_DOTPROD(3);
	    COMPUTE_DOTPROD(4);
	    COMPUTE_DOTPROD(5);
	    COMPUTE_DOTPROD(6);
	    COMPUTE_DOTPROD(7);
	    
	    


#define COMPUTE_INDEX(idx) \
	    tmp_cmps_a[0]= spu_cmpgt(dot_v[2*idx], binb_dv[nbins]);  \
	    tmp_cmps_b[0] =spu_cmpgt(dot_v[2*idx+1], binb_dv[nbins]);	\
	    cmps[0] = (vector unsigned int)spu_shuffle(tmp_cmps_a[0], tmp_cmps_b[0], cmp_shuf_v);	    \
	    tmp_cmps_a[1]= spu_cmpgt(dot_v[2*idx], binb_dv[nbins-1]);	\
	    tmp_cmps_b[1]= spu_cmpgt(dot_v[2*idx+1], binb_dv[nbins-1]);  \
	    cmps[1] = (vector unsigned int)spu_shuffle(tmp_cmps_a[1], tmp_cmps_b[1], cmp_shuf_v); \
	    tmp_cmps_a[2]= spu_cmpgt(dot_v[2*idx], binb_dv[nbins-2]);  \
	    tmp_cmps_b[2]= spu_cmpgt(dot_v[2*idx+1], binb_dv[nbins-2]);  \
	    cmps[2] = (vector unsigned int)spu_shuffle(tmp_cmps_a[2], tmp_cmps_b[2], cmp_shuf_v); \
	    tmp_cmps_a[3]= spu_cmpgt(dot_v[2*idx], binb_dv[nbins-3]);  \
	    tmp_cmps_b[3]= spu_cmpgt(dot_v[2*idx+1], binb_dv[nbins-3]);  \
	    cmps[3] = (vector unsigned int)spu_shuffle(tmp_cmps_a[3], tmp_cmps_b[3], cmp_shuf_v); \
	    tmp_cmps_a[4]= spu_cmpgt(dot_v[2*idx], binb_dv[nbins-4]);  \
	    tmp_cmps_b[4]= spu_cmpgt(dot_v[2*idx+1], binb_dv[nbins-4]); \
	    cmps[4] = (vector unsigned int)spu_shuffle(tmp_cmps_a[4], tmp_cmps_b[4], cmp_shuf_v); \
	    tmp_cmps_a[5]= spu_cmpgt(dot_v[2*idx], binb_dv[nbins-5]);  \
	    tmp_cmps_b[5]= spu_cmpgt(dot_v[2*idx+1], binb_dv[nbins-5]);  \
	    cmps[5] = (vector unsigned int)spu_shuffle(tmp_cmps_a[5], tmp_cmps_b[5], cmp_shuf_v); \
	    tmp_cmps_a[6]= spu_cmpgt(dot_v[2*idx], binb_dv[nbins-6]);  \
	    tmp_cmps_b[6]= spu_cmpgt(dot_v[2*idx+1], binb_dv[nbins-6]);  \
	    cmps[6] = (vector unsigned int)spu_shuffle(tmp_cmps_a[6], tmp_cmps_b[6], cmp_shuf_v); \
	    tmp_cmps_a[7]= spu_cmpgt(dot_v[2*idx], binb_dv[nbins-7]);  \
	    tmp_cmps_b[7]= spu_cmpgt(dot_v[2*idx+1], binb_dv[nbins-7]); \
	    cmps[7] = (vector unsigned int)spu_shuffle(tmp_cmps_a[7], tmp_cmps_b[7], cmp_shuf_v); \
	    dotgather[idx] = spu_gather(cmps[7]); \
	    cmps[0] = spu_and(cmps[0], cmpmask); \
	    cmps[1] = spu_and(cmps[1], cmpmask); \
	    cmps[2] = spu_and(cmps[2], cmpmask); \
	    cmps[3] = spu_and(cmps[3], cmpmask); \
	    cmps[4] = spu_and(cmps[4], cmpmask); \
	    cmps[5] = spu_and(cmps[5], cmpmask); \
	    cmps[6] = spu_and(cmps[6], cmpmask); \
	    cmps[7] = spu_and(cmps[7], cmpmask); \
	    t0 = spu_add(cmps[0], cmps[1]);  \
	    t1 = spu_add(cmps[2], cmps[3]);  \
	    t2 = spu_add(cmps[4], cmps[5]);  \
	    t3 = spu_add(cmps[6], cmps[7]);  \
	    t4 = spu_add(t0, t1);  \
	    t5 = spu_add(t2, t3);  \
	    t6 = spu_add(t4, t5);  \
	    local_sel_v = spu_cmpgt(t6, i_seven_v); \
	    t6 = spu_sel(t6, i_zero_v, local_sel_v); \
	    adjusted_index[idx]=t6;
	    
	    COMPUTE_INDEX(0);
	    COMPUTE_INDEX(1);
	    COMPUTE_INDEX(2);
	    COMPUTE_INDEX(3);
	    COMPUTE_INDEX(4);
	    COMPUTE_INDEX(5);
	    COMPUTE_INDEX(6);
	    COMPUTE_INDEX(7);



#define COMPUTE_REST(idx) \
	    index_array[index_array_count] = adjusted_index[idx];	\
	    index_array_count++;		     \
	    if (spu_extract(dotgather[idx], 0) != 0){ \
		test_compute_smallp(dot_v[2*idx],dot_v[2*idx+1], dotgather[idx], jk); \
	    }
	    
	    
	    COMPUTE_REST(0);
	    COMPUTE_REST(1);
	    COMPUTE_REST(2);
	    COMPUTE_REST(3);
	    COMPUTE_REST(4);
	    COMPUTE_REST(5);
	    COMPUTE_REST(6);
	    COMPUTE_REST(7);
	    
	    
	}//jj for 32
	
	
	for(; jj<ROUNDDOWN_4(blocksize_j); jj+=4){
	    vector double* m = (vector double*)(&d2[jj]);
	    COMPUTE_DOTPROD(0);
	    COMPUTE_INDEX(0);
	    COMPUTE_REST(0);
	    
	} //jj for 4
	
	if(jj<blocksize_j){
	    vector float* m = (vector float*)(&d2[jj]);
	    // xxxx yyyy zzzz tttt
	    struct cartesian boundary_d2[4];
	    convert_back(m, (vector float*)boundary_d2);	    
	    int kk;
	    for(kk=0; kk< blocksize_j - jj; kk++){
		compute_one_scalar(&d1[ii], &boundary_d2[kk]);
	    }
	    
	}

	
	partial_compute_dot(jk);
	
	vector unsigned short subtotal = compute_subtotal(index_array_count);
	vector unsigned short subtotal2 = compute_subtotal2(index_array2_count);

	temp_data_bins[nbins+1] += spu_extract(subtotal, 0);
	temp_data_bins[nbins]   += spu_extract(subtotal, 1);
	temp_data_bins[nbins-1] += spu_extract(subtotal, 2);
	temp_data_bins[nbins-2] += spu_extract(subtotal, 3);
	temp_data_bins[nbins-3] += spu_extract(subtotal, 4);
	temp_data_bins[nbins-4] += spu_extract(subtotal, 5);
	temp_data_bins[nbins-5] += spu_extract(subtotal, 6);
	temp_data_bins[nbins-6] += spu_extract(subtotal, 7);

	temp_data_bins[nbins-7] += spu_extract(subtotal2, 0);
	temp_data_bins[nbins-8] += spu_extract(subtotal2, 1);
	temp_data_bins[nbins-9] += spu_extract(subtotal2, 2);
	temp_data_bins[nbins-10] += spu_extract(subtotal2, 3);
	temp_data_bins[nbins-11] += spu_extract(subtotal2, 4);
	temp_data_bins[nbins-12] += spu_extract(subtotal2, 5);
	temp_data_bins[nbins-13] += spu_extract(subtotal2, 6);
	temp_data_bins[nbins-14] += spu_extract(subtotal2, 7);
	
	if (jk >= 0){
	    data_bins[jk][nbins+1] -= spu_extract(subtotal, 0);
	    data_bins[jk][nbins] -= spu_extract(subtotal, 1);
	    data_bins[jk][nbins-1] -= spu_extract(subtotal, 2);
	    data_bins[jk][nbins-2] -= spu_extract(subtotal, 3);
	    data_bins[jk][nbins-3] -= spu_extract(subtotal, 4);
	    data_bins[jk][nbins-4] -= spu_extract(subtotal, 5);
	    data_bins[jk][nbins-5] -= spu_extract(subtotal, 6);		
	    data_bins[jk][nbins-6] -= spu_extract(subtotal, 7);

	    data_bins[jk][nbins-7] -= spu_extract(subtotal2, 0);
	    data_bins[jk][nbins-8] -= spu_extract(subtotal2, 1);
	    data_bins[jk][nbins-9] -= spu_extract(subtotal2, 2);
	    data_bins[jk][nbins-10] -= spu_extract(subtotal2, 3);
	    data_bins[jk][nbins-11] -= spu_extract(subtotal2, 4);
	    data_bins[jk][nbins-12] -= spu_extract(subtotal2, 5);
	    data_bins[jk][nbins-13] -= spu_extract(subtotal2, 6);		
	    data_bins[jk][nbins-14] -= spu_extract(subtotal2, 7);
	    
	}

    }//ii
    
}


void
display_results(void)
{
    int i, j;
    for (i = 0;i < 11; i++){
	for (j =0 ; j < MAXBINS; j++){
	    printf("%lld  ", data_bins[i][j]);	    
	}
	printf("\n");
    }
    
}

int
compute_one_scalar(struct cartesian* p0, struct cartesian* p1)
{
    int jk = p0->jk;	    
	    
    int indx;
    double dot = p0->x*p1->x + p0->y*p1->y + p0->z*p1->z;    
    int k = nbins;
    int l;
    
    if (dot >= binb_d[0])
	indx = 0;
    else
	{
	    while (dot > binb_d[k]){
		k--;
	    }
	    indx = k+1;
	}
    
    for(l=0; l<njk; l++){
	if(l!=jk){
	    data_bins[l][indx] +=1;
	}
    }
    
    return 0;
    
}

int
compute_bins_auto(compute_task_t* task)
{
    
    int i, j;
    int ii, jj, k, l;

    int tag_id = 0;
    
    int n1 =  task->n1;
    int n2 = task->n2;
    
    memset(temp_data_bins, 0, sizeof(temp_data_bins));

    
    DODMA(data_bins, task->data_bins, MAXBINS*MAXNJK*sizeof(long long), tag_id, MFC_GET_CMD);
    spu_writech(MFC_WrTagMask, 1<<tag_id);
    spu_mfcstat(2);    
    

    //spu_log("before auto computing\n");
    //display_results();

    int blocksize_i;
    int blocksize_j;
    
    for ( i =0;i < n1-1 ; i+= blocksize_i){
	blocksize_i =  (n1 - i )>=BLOCK ? BLOCK : (n1 - i);
	
	DODMA(data1, task->data1 + sizeof(struct cartesian)* i, 
	      sizeof(struct cartesian)*ROUNDUP_4(blocksize_i), tag_id, MFC_GET_CMD);
	DODMA(data2, task->data2_v + sizeof(struct cartesian)* i, 
	      sizeof(struct cartesian)*ROUNDUP_4(blocksize_i), tag_id, MFC_GET_CMD);
	spu_writech(MFC_WrTagMask, 1<<tag_id);
	spu_mfcstat(2);
	
	
	for(ii=0; ii<blocksize_i; ii+=4){
	    
	    compute_one_scalar(&data1[ii], &data1[ii+1]);
	    compute_one_scalar(&data1[ii], &data1[ii+2]);
	    compute_one_scalar(&data1[ii], &data1[ii+3]);
	    compute_one_scalar(&data1[ii+1], &data1[ii+2]);
	    compute_one_scalar(&data1[ii+1], &data1[ii+3]);
	    compute_one_scalar(&data1[ii+2], &data1[ii+3]);
   	    
	    block_compute(4, blocksize_i - (ii + 4), &data1[ii], &data2[ii+4]);
	    
	}//ii
	
	
	for (j=i + blocksize_i; j < n2; j+= blocksize_j){
	    blocksize_j = (n2 -j) >= BLOCK? BLOCK: (n2 - j);
	    DODMA(data2, task->data2_v + sizeof(struct cartesian)*j , 
		  sizeof(struct cartesian)*ROUNDUP_4(blocksize_j), tag_id, MFC_GET_CMD);
	    spu_writech(MFC_WrTagMask, 1<< tag_id);
	    spu_mfcstat(2);
	    
	    block_compute(blocksize_i, blocksize_j, data1, data2);
	    
	}//j		
	


    } //i
    

    long long total_0_23 = 0;
    
    for (j =0; j < nbins+2-8; j++){
	total_0_23 += temp_data_bins[j];
    }
    
    temp_data_bins[nbins+1] -=  total_0_23;
    

    
    for (i=0;i < task->njk;i++){
	for (j =0; j < nbins+2; j++){
	    data_bins[i][j] += temp_data_bins[j];
	}
    }

    
    DODMA(data_bins, task->data_bins, sizeof(long long)*MAXBINS*MAXNJK, tag_id, MFC_PUT_CMD);
    spu_writech(MFC_WrTagMask, 1<<tag_id);
    spu_mfcstat(2);    
    
    //spu_log("after auto computing\n");
    //display_results();
    return 0;
}

int 
compute_bins_cross(compute_task_t* task)
{
    int i, j;
    int tag_id = 0;
    
    int n1 =  task->n1;
    int n2 = task->n2;
    
    memset(temp_data_bins, 0, sizeof(temp_data_bins));
    DODMA(data_bins, task->data_bins, MAXBINS*MAXNJK*sizeof(long long), tag_id, MFC_GET_CMD);
    spu_writech(MFC_WrTagMask, 1<<tag_id);
    spu_mfcstat(2);    
    
    int blocksize_i;
    int blocksize_j;

    for ( i =0;i < n1 ; i+= blocksize_i){
	blocksize_i =  (n1 - i )>=BLOCK ? BLOCK : (n1 - i);
	DODMA(data1, task->data1 + sizeof(struct cartesian)*i, 
	      sizeof(struct cartesian)*ROUNDUP_4(blocksize_i), tag_id, MFC_GET_CMD);
	spu_writech(MFC_WrTagMask, 1<<tag_id);
	spu_mfcstat(2);
	
	for (j=0; j < n2; j+=blocksize_j){
	    blocksize_j = (n2 -j) >= BLOCK? BLOCK: (n2 - j);
	    DODMA(data2, task->data2_v + sizeof(struct cartesian)* j, 
		  sizeof(struct cartesian)*ROUNDUP_4(blocksize_j), tag_id, MFC_GET_CMD);
	    spu_writech(MFC_WrTagMask, 1<< tag_id);
	    spu_mfcstat(2);
	    
	    block_compute(blocksize_i, blocksize_j, data1, data2);
	    
	    
	}//j		
    } //i


    //we have over-computed the last bins
    //we must substract them from the last bin, the last bin's index is (nbins+1)
    long long total_0_23 = 0;
    
    for (j =0; j < nbins+2-8; j++){
	total_0_23 += temp_data_bins[j];
    }
    temp_data_bins[nbins+1] -=  total_0_23;
   
    for (i=0;i < njk;i++){
	for (j =0; j < nbins+2; j++){
	    data_bins[i][j] += temp_data_bins[j];
	}
    }
    
    
   
    DODMA(data_bins, task->data_bins, MAXNJK*MAXBINS*sizeof(long long), tag_id, MFC_PUT_CMD);
    spu_writech(MFC_WrTagMask, 1<<tag_id);
    spu_mfcstat(2);    
    
    
    return 0;
    
    
}

int 
compute_bins(task_t* _task)
{
    compute_task_t* task = (compute_task_t*)_task;
    njk = task->njk;
    
    if (task->doSelf){
	compute_bins_auto(task);
    }else{
	compute_bins_cross(task);
    }

    return 0;    
    
}

